//
//  PuzzleView.m
//  puzzle
//
//  Created by kunii on 11/06/18.
//  Copyright 2011 __MyCompanyName__. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>
#import <ImageIO/ImageIO.h>
#import <MobileCoreServices/MobileCoreServices.h>
#import "PuzzleView.h"

/*
 *  パズル用タイルクラス
 */
@interface PuzzleTileLayer : CALayer {
    /// 画像インディックス
    int number;
}
/// 自身がドラッグ中はYES
@property (assign) BOOL dragging;
@property (assign) int locationNumber;
@end
@implementation PuzzleTileLayer
@synthesize dragging, locationNumber;
/// 横分割数
static int divH;
/// 縦分割数
static int divV;

/**
 *  分割数の設定
 */
+ (void)setDivH:(int)h v:(int)v
{
    divH = h;
    divV = v;
}

/*
 *  インディックスの記憶と、インディックスにあったcontentsRectの設定
 */
- (void)setNumber:(int)inNumber
{
    number = inNumber;
}

- (void)makeContent:(UIImage*)image
{
    CGSize size = image.size;
    int width = size.width / divH;
    int height = size.height / divV;
    UIGraphicsBeginImageContext(CGSizeMake(width, height));
    int h = number % divH;
    int v = number / divH;
    [image drawAtPoint:CGPointMake(-h * width, -v * height)];
    image = UIGraphicsGetImageFromCurrentImageContext();
    UIGraphicsEndImageContext();
    self.contents = (id)image.CGImage;    
}

- (UIImage*)image
{
    return [UIImage imageWithCGImage:(CGImageRef)self.contents];
}

-(CGPoint)location
{
    int width = CGImageGetWidth((CGImageRef)self.contents);
    int height = CGImageGetHeight((CGImageRef)self.contents);
    int h = locationNumber % divH;
    int v = locationNumber / divH;
    return CGPointMake(h * width, v * height);
}

/*
 *  インディックスを返す
 */
-(int)number
{
    return number;
}

/*
 *  選択中にする
 */
- (void)setSelected:(BOOL)selected
{
    if (selected) {
        [self retain];
        CALayer* super_layer = self.superlayer;
        [self removeFromSuperlayer];
        [super_layer addSublayer:self];
        [self release];
        
        self.shadowRadius = 8;
        self.shadowOffset = CGSizeMake(0, 0);
        self.shadowColor = [UIColor blackColor].CGColor;
        self.shadowOpacity = 1.0;    

        CABasicAnimation* prompt = [CABasicAnimation animationWithKeyPath:@"transform"];
        prompt.duration = 0.5;
        prompt.toValue = [NSValue valueWithCATransform3D:CATransform3DMakeScale(1.1, 1.1, 1.0)];
        prompt.repeatCount = HUGE_VALF;
        prompt.autoreverses = YES;
        [self addAnimation:prompt forKey:@"blink"];    
    } else {
        self.shadowOpacity = 0.0;
        [self removeAnimationForKey:@"blink"];
    }
}

/**
 *  ぶるぶる
 */
- (void)setVibration:(BOOL)on
{
    if (on) {
        CABasicAnimation* prompt = [CABasicAnimation animationWithKeyPath:@"transform"];
        prompt.duration = 0.1;
        prompt.fromValue = [NSValue valueWithCATransform3D:CATransform3DMakeRotation(-0.05, 0.0, 0.0, 1.0)];
        prompt.toValue = [NSValue valueWithCATransform3D:CATransform3DMakeRotation(0.05, 0.0, 0.0, 1.0)];
        prompt.repeatCount = HUGE_VALF;
        prompt.autoreverses = YES;
        [self addAnimation:prompt forKey:@"vibration"];    
    } else {
        [self removeAnimationForKey:@"vibration"];
    }    
}

- (CALayer*)hitTest:(CGPoint)p
{
    if (dragging)
        return nil;
    return [super hitTest:p];
}
@end

@implementation PuzzleView

- (id)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        // Initialization code
    }
    return self;
}

/*
// Only override drawRect: if you perform custom drawing.
// An empty implementation adversely affects performance during animation.
- (void)drawRect:(CGRect)rect
{
    // Drawing code
}
*/

- (void)dealloc
{
    [super dealloc];
}

/**
 *  ぶるぶる
 */
- (void)vibration:(BOOL)on
{
    for (CALayer* layer in self.layer.sublayers) {
        if ([layer isKindOfClass:[PuzzleTileLayer class]]) {
            [(PuzzleTileLayer*)layer setVibration:on];
        }
    }
}

/*
 最適サイズ画像作成
 */
- (UIImage*)optimumImage:(UIImage*)image withHCount:(int)hCount vCount:(int)vCount
{
    CGSize size = image.size;
    int optimumWidth = ((int)size.width / hCount) * hCount;
    int optimumHeight = ((int)size.height / vCount) * vCount;
    UIGraphicsBeginImageContext(CGSizeMake(optimumWidth, optimumHeight));
    int left = (int)(size.width - optimumWidth) / 2;
    int top = (int)(size.height - optimumHeight) / 2;
    [image drawAtPoint:CGPointMake(left, top)];
    image = UIGraphicsGetImageFromCurrentImageContext();
    UIGraphicsEndImageContext();
    return image;
}

- (UIImage*)curtImage
{
    UIGraphicsBeginImageContext(_image.size);
    for (PuzzleTileLayer* layer in self.layer.sublayers) {
        [[layer image] drawAtPoint:[layer location]];
    }
    UIImage* image = UIGraphicsGetImageFromCurrentImageContext();
    UIGraphicsEndImageContext();
    return image;    
}

- (void)setImage:(UIImage*)image withHCount:(int)hCount vCount:(int)vCount
{
    image = [self optimumImage:image withHCount:hCount vCount:vCount];
    [_image release];
    _image = [image retain];
    //  縦横の分割数を保存
    horizontalCount = hCount;
    verticalCount = vCount;
    //  以前のタイルを全削除
    NSArray* array = [self.layer.sublayers copy];
    for (CALayer* layer in array) {
        [layer removeFromSuperlayer];
    }
    //  新しいタイル1枚（左上）の矩形を決定
    CGRect r = CGRectZero;
    r.size.width = self.bounds.size.width / (float)horizontalCount;
    r.size.height = self.bounds.size.height / (float)verticalCount;
    //  タイルクラスに分割数を設定
    [PuzzleTileLayer setDivH:horizontalCount v:verticalCount];
    for (int v = 0; v < verticalCount; v++) {
        for (int h = 0; h < horizontalCount; h++) {
            PuzzleTileLayer* tile = [PuzzleTileLayer layer];
            tile.frame = r;
            [tile setNumber:v * horizontalCount + h];
            tile.locationNumber = tile.number;
            [tile makeContent:_image];
            [self.layer addSublayer:tile];
            r = CGRectOffset(r, r.size.width, 0);   //  次の位置へ
        }
        //  矩形を左端に戻し、縦位置を一段ずらず
        r.origin.x = self.bounds.origin.x;  
        r = CGRectOffset(r, 0, r.size.height);
    }
    [self vibration:YES];
}

/**
 *  タイル差し替え
 *      fromTileとtileを差し替える
 */
-(void)swapTile:(PuzzleTileLayer*)tile
{
    [CATransaction begin];
    [CATransaction setCompletionBlock:^{
        int locationNumber = tile.locationNumber;
        tile.locationNumber = fromTile.locationNumber;
        fromTile.locationNumber = locationNumber;
        [fromTile setSelected:NO];
        fromTile = nil;        
    }];
    //  fromTileをtileに移動するアニメーション
    CABasicAnimation* move = [CABasicAnimation animationWithKeyPath:@"position"];
    move.toValue = [NSValue valueWithCGPoint:tile.position];
    [fromTile addAnimation:move forKey:@"move"];
    //  tileをfromTileに移動するアニメーション
    move = [CABasicAnimation animationWithKeyPath:@"position"];
    move.toValue = [NSValue valueWithCGPoint:fromTile.position];
    [tile addAnimation:move forKey:@"move"];
    [CATransaction commit];
}

/**
 *  タップされたタイルを特定し、1つめなら選択、２つめなら最初のタイルと差し替える
 */
- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
    [self vibration:NO];
    //  マルチフィンガー指定はしていないので、anyObjectで取り出し。
    UITouch* touch = [touches anyObject];    
    //  タップ位置を自分のローカル座標に変換
    CGPoint pt = [touch locationInView:self];
    //  ただし、CALayerのhitoTestには親CALayer上での位置が必要なので、もう一回変換
    CGPoint superpt = [self.layer convertPoint:pt toLayer:self.layer.superlayer];
    PuzzleTileLayer* tile = (PuzzleTileLayer*)[self.layer hitTest:superpt];
    if (tile == nil) {
        //  どのタイルにも触れていないならおしまい
        return;
    }
    fromTile = tile;
    fromTile.dragging = YES;
    fromTilePosition = fromTile.position;
    offset.x = fromTile.position.x - pt.x;
    offset.y = fromTile.position.y - pt.y;
    toTile = nil;
}

-(void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
    //  マルチフィンガー指定はしていないので、anyObjectで取り出し。
    UITouch* touch = [touches anyObject];    
    //  タップ位置を自分のローカル座標に変換
    CGPoint pt = [touch locationInView:self];
    //  ドラッグに合わせてfromTileを移動
    [CATransaction begin];
    [CATransaction setDisableActions:YES];
    fromTile.position = CGPointMake(offset.x + pt.x, offset.y + pt.y);
    [CATransaction commit];    

    //  ただし、CALayerのhitoTestには親CALayer上での位置が必要なので、もう一回変換
    CGPoint superpt = [self.layer convertPoint:pt toLayer:self.layer.superlayer];
    PuzzleTileLayer* hitTile = (PuzzleTileLayer*)[self.layer hitTest:superpt];
    if ((hitTile == nil) || (hitTile == self.layer)) {
        return;
    }
    if (toTile != hitTile) {
        if (toTile) {
            //  古い差し替えタイルがあれば、元に戻す
            toTile.position = toTilePosition;
        }
        //  差し替えタイルの元のpositionを保存してからfromTilePositionに設定
        toTile = hitTile;
        toTilePosition = toTile.position;
        toTile.position = fromTilePosition;
    }
}

/**
 *  ドラッグ終了
 */
-(void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
    if (toTile) {
        fromTile.position = toTilePosition;
        int locationNumber = toTile.locationNumber;
        toTile.locationNumber = fromTile.locationNumber;
        fromTile.locationNumber = locationNumber;
    }
    fromTile.dragging = NO;
    fromTile = nil;
    [self vibration:YES];
}

/**
 *  電話等で割り込みが入ったので、選択タイルがあれば解除する
 */
-(void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
{
    fromTile.position = fromTilePosition;
    toTile.position = toTilePosition;
    fromTile.dragging = NO;
    fromTile = nil;
    [self vibration:YES];
}

- (void)export:(NSString*)path {
	NSMutableData* data = [[[NSMutableData alloc] init] autorelease];
	CGImageDestinationRef destref = CGImageDestinationCreateWithData((CFMutableDataRef)data, kUTTypeTIFF, 1, nil);

    //  1024バイトの制限有り    16 x 16分割程度が限界となる
    NSArray* array = [self.layer.sublayers sortedArrayUsingComparator:^(id obj1, id obj2) {
        PuzzleTileLayer* l1 = obj1;
        PuzzleTileLayer* l2 = obj2;
        if (l1.locationNumber < l2.locationNumber)
            return NSOrderedAscending;
        if (l1.locationNumber > l2.locationNumber)
            return NSOrderedDescending;
        return NSOrderedSame;
    }];
    NSString* description = [NSString stringWithFormat:@"PUZZLE,%03d,%03d", divH, divV];
    for (PuzzleTileLayer* layer in array) {
        description = [description stringByAppendingFormat:@",%03d", layer.number];
    }
    
    NSLog(@"%@", description);
    NSDictionary* tifprop = [NSDictionary dictionaryWithObjectsAndKeys:description, (id)kCGImagePropertyTIFFImageDescription, nil];
    NSDictionary*  prop = [NSDictionary dictionaryWithObjectsAndKeys:tifprop, (id)kCGImagePropertyTIFFDictionary, nil];
	CGImageDestinationAddImage(destref, [self curtImage].CGImage, (CFDictionaryRef)prop);
	CGImageDestinationFinalize(destref);
	[data writeToFile:path atomically:YES];
	CFRelease(destref);
}

- (BOOL)import:(NSString*)path {
	CGImageSourceRef srcref = CGImageSourceCreateWithURL((CFURLRef)[NSURL fileURLWithPath:path], nil);
	NSDictionary *dic = (NSDictionary *)CGImageSourceCopyPropertiesAtIndex(srcref, 0, nil);
    NSDictionary *tifprop = [dic objectForKey:(id)kCGImagePropertyTIFFDictionary];
    NSString *commentStr = [tifprop objectForKey:(id)kCGImagePropertyTIFFImageDescription];
    NSLog(@"%@", commentStr);
    NSArray* locations = [commentStr componentsSeparatedByString:@","];
	[dic release];
	CFRelease(srcref);
    
    if ([@"PUZZLE" isEqualToString:[locations objectAtIndex:0]]== NO) {
        return NO;
    }
    int hCount = [[locations objectAtIndex:1] intValue];
    int vCount = [[locations objectAtIndex:2] intValue];
    UIImage* image = [UIImage imageWithContentsOfFile:path];
    image = [self optimumImage:image withHCount:hCount vCount:vCount];
    [_image release];
    _image = [image retain];
    //  縦横の分割数を保存
    horizontalCount = hCount;
    verticalCount = vCount;
    //  以前のタイルを全削除
    NSArray* array = [self.layer.sublayers copy];
    for (CALayer* layer in array) {
        [layer removeFromSuperlayer];
    }
    //  新しいタイル1枚（左上）の矩形を決定
    CGRect r = CGRectZero;
    r.size.width = self.bounds.size.width / (float)horizontalCount;
    r.size.height = self.bounds.size.height / (float)verticalCount;
    //  タイルクラスに分割数を設定
    [PuzzleTileLayer setDivH:horizontalCount v:verticalCount];
    for (int v = 0; v < verticalCount; v++) {
        for (int h = 0; h < horizontalCount; h++) {
            PuzzleTileLayer* tile = [PuzzleTileLayer layer];
            tile.frame = r;
            tile.number = v * horizontalCount + h;
            tile.locationNumber = tile.number;
            [tile makeContent:_image];
            tile.number = [[locations objectAtIndex:tile.number + 3] intValue];
            [self.layer addSublayer:tile];
            r = CGRectOffset(r, r.size.width, 0);   //  次の位置へ
        }
        //  矩形を左端に戻し、縦位置を一段ずらず
        r.origin.x = self.bounds.origin.x;  
        r = CGRectOffset(r, 0, r.size.height);
    }
    [self vibration:YES];
    return YES;
}
@end
